package PureMessage::WebUI::Controller::LogSearch;
use strict;
use warnings;
use URI::Escape;
use HTTP::Date;
use POSIX qw(strftime mktime);
#use ActiveState::I18N::UTF8 qw(utf8_header);
#use PureMessage::UTF8::Config qw(%utf8_config);
use PureMessage::QDigest::MessageId qw(extract_ids decode_id);
use PureMessage::WebUI::Utils qw(htmlify get_message_status $DEFAULT_DATE
                                 message_status_from_id mset_query
                                 validate_start_end_dates parse_date);
use PureMessage::Group::PermissionItem;
use Sophos::LogSearch;
use Sophos::LogSearch::Constants qw(utc_time_string_to_epoch $MAX_RESULTS);
use POSIX qw(strftime);
use Date::Parse qw(strptime);
use Time::Local qw(timelocal timegm);

use base 'Exporter';
our @EXPORT_OK = qw(handle_log_search handle_logsearch_preview);

sub remove_extra_white_space {
    my $data = shift;
    return $data if !defined $data;
    $data =~ s/^\s+//g; # remove whitespace from the beginning and the end
    $data =~ s/\s+$//g;
    return $data;
}

sub handle_log_search {
    my ($app) = @_;
    my $context = $app->{context};
    
    my $s = Sophos::LogSearch->new();
    my @log_entries; 

    my @msg;
    my %default_args = (
        offset => 0,
        count  => 50,
        hosts  => '*',
    );
    
    my %args = (%default_args);

    for my $key (qw(from msgid action reason to mailroute page_size subject)){
    #for my $key (qw(from recipient subject reason dateFrom dateTo relay queue_id)){
        my $temp = $app->query->param($key);
        $temp = remove_extra_white_space($temp);
        $context->{search}{params}{$key} = $temp;
        $temp =~ s/_/__/g if ($key eq 'to' ); # Workaround if we are searching the recipient with _
        $temp =~ s/\s/^^/g if ($key eq 'subject' && $temp);
        if ($temp) {
            push(@log_entries,"$key='$temp'") if !($key eq q{page_size}); 
            $args{$key} = Sophos::LogSearch::clean_and_escape($temp);
            $args{$key} =~ s/\*/.*/g;
            $args{count} = $args{$key} if ($key eq q(page_size));
        }
    }
    # aliases contains both the envelope to and the mapped recipients, for purposes of 
    # searching we will use it instead of the to address... make the switch here to
    # present to the search api
    $args{aliases} = delete($args{to}) if $args{to};

    my $search_text = scalar(@log_entries) ? join(' ', sort(@log_entries)) : "''";
    $app->log("logsearch: " . $search_text);

    # Add groups to restrict the search
    my $user = $context->{user};
    if(!$user->super_user){
        # XXX: hack to overcome cdbi's stupid use of overload
        $args{groups} = [
            map { my $g = '' . $_->group_name; $g =~ s/_/__/g; $g } $user->groups
        ];
    }
    
    my $loc = $context->{loc};

    # handle date input.  Dates are expected to be in a modified ISO format in localtime
    # eg.  2007-09-03[08:15]
    # leading zero's are required!
      
    my $start_date = remove_extra_white_space($app->query->param('ldateFrom')); # time string in local time from user, format 'yyyy-mm-dd hh:mm'
    my $end_date = remove_extra_white_space($app->query->param('ldateTo'));
     
    $start_date = ($start_date && $start_date eq $DEFAULT_DATE) ? '' : Sophos::LogSearch::clean_and_escape($start_date);
    $end_date = ($end_date && $end_date eq $DEFAULT_DATE) ? '' : Sophos::LogSearch::clean_and_escape($end_date);

    $context->{note} = '';
    if ($start_date or $end_date) {
        # magic number 946684800 = 2001-01-01T00:00:00
        $end_date = format_epoch(time) unless $end_date;
        my ($f_end_date,$d_end_date) = guess_date($end_date,1);
        my ($f_start_date,$d_start_date);
        
        if ($f_end_date and !$start_date) {
          my $end_epoch = str2time($f_end_date,'UTC');
          $end_epoch -= (60 * 60* 24 * 30); # 30 days earlier
          $start_date = format_epoch($end_epoch);
          
          ($f_start_date,$d_start_date) = guess_date($start_date);

        } elsif($start_date) {

          ($f_start_date,$d_start_date) = guess_date($start_date);
        
        } else {

          $f_start_date = format_epoch(time());
          $d_start_date = $DEFAULT_DATE;
        }


        if (not ($f_start_date and $f_end_date) ) {
            $context->{statusbar} = 'bad';
            my $bad_date = $f_start_date ? $end_date : $start_date;
            #Replace underscore separator with spaces
            $bad_date =~ s/_/ /g;

            $context->{note} = $loc->('Invalid date: [quant,_1]',$bad_date);
            $context->{search}{params}{ldateFrom} = $f_start_date ? $d_start_date : $start_date;
            $context->{search}{params}{ldateTo} = $f_end_date ? $d_end_date :  $end_date;

            #Replace underscore separator with spaces
            $context->{search}{params}{ldateFrom} =~ s/_/ /g;
            $context->{search}{params}{ldateTo}   =~ s/_/ /g;
            return;
        }
       
        my $utc_delta = str2time($d_end_date,'UTC') - str2time($d_start_date,'UTC');
        my $local_delta = str2time($d_end_date) - str2time($d_start_date);
       
        if($utc_delta != $local_delta) {
           #if( ( str2time($f_end_date) > str2time($f_start_date) ) and (str2time($d_end_date) - str2time($d_start_date)  )) {
          # then there was a tz conversion issue, likely over a daylight savings time boundry # oct 14
          $context->{note} = $loc->('Your date range crosses a time change, results may not be as expected.');
           
        }

        $args{atime} = $f_start_date;
        $args{ztime} = $f_end_date;
        $args{atime} =~ s/ /_/g; # log search needs to have spaces re-encoded into _'s 
        $args{ztime} =~ s/ /_/g;
        #Replace underscore separator with spaces
        $d_start_date =~ s/_/ /g;
        $d_end_date   =~ s/_/ /g;
        $context->{search}{params}{ldateFrom} = $d_start_date;
        $context->{search}{params}{ldateTo}   = $d_end_date;
    } else {
        my ($f_start_date,$d_start_date, $f_end_date, $d_end_date);
        ($f_start_date,$d_start_date) = guess_date(_earliest_datetime(time()), 0);
        ($f_end_date,$d_end_date) = guess_date(_latest_datetime(time()), 1);
        $args{atime} = $f_start_date;
        $args{ztime} = $f_end_date;
        $args{atime} =~ s/ /_/g; # log search needs to have spaces re-encoded into _'s 
        $args{ztime} =~ s/ /_/g;
        #Replace underscore separator with spaces
        $d_start_date =~ s/_/ /g;
        $d_end_date   =~ s/_/ /g;
        $context->{search}{params}{ldateFrom} = $d_start_date;
        $context->{search}{params}{ldateTo}   = $d_end_date;
    }

    my $cnt = 1;
    my $more_than_max = 0;
    foreach my $message (@{ $s->search(%args) }){
        if ($cnt > $MAX_RESULTS) {
            # if result is MAX_RESULTS + 1, notify user to refine date
            # otherwise result shows badly sorted  MAX_RESULTS (DEF24389)
            $more_than_max = 1;
            last;
        }
        my %data;
        $data{$_} = $message->field($_) for qw( id action reason from to subject);
        $data{subject} =~ s/\^\^/ /g;
        $data{reason} =~ s/_/ /g;
        $data{time} = strftime('%Y-%m-%dT%H:%M:%S', localtime( utc_time_string_to_epoch($message->field('time'))) ) ; # convert UTC to localtime
        $data{to} =~ s/__/_/g;
        $data{to} = join(', ', split(',', $data{to}));

        for (qw(action reason)){
            $data{$_} = $data{$_} =~ m{,} ? $loc->('Multi') : $data{$_}; 
        }

        # add a fetchkey so we can get a hold of this particular message again
        $data{fetchkey} = $message->field('fetchkey');

        push(@msg,\%data);
        ++$cnt;
    }


    if($s->http_errors()){
        my ($level,$errors) = @{ $s->http_errors() };
        if($level eq 'error'){
            $context->{statusbar} = q{bad};
            $context->{note} = $loc->("Unable to search remote servers");
            my $log_text = "logsearch: Error occured during remote search:\n";
            for my $e (@{$errors}){
                $log_text .= "\t" . $e->{url} . ": \n";
                $log_text .= "\t  ". $e->{message} . "\n";
            }
            $app->log($log_text);
            $context->{search}{results} = \@msg;
        
        }elsif($level eq 'warning'){
            $context->{statusbar} = q{warning};
            $context->{search}{results} = \@msg;
            my $log_text = "logsearch: Error occured during remote search:\n";
            my @err_hosts;
            for my $e (@{$errors}){
                $log_text .= "\t" . $e->{url} . ": \n";
                $log_text .= "\t  ". $e->{message} . "\n";
                push(@err_hosts, $e->{host});
            }
            my $err_host_list = join(", ", @err_hosts);
            $context->{note} = $loc->("The following remote servers are not running HTTP(RPC/UI): $err_host_list");
            $app->log($log_text);
            $context->{search}{results} = \@msg;
        
        }

    }else {
        $context->{statusbar} = q{good};
        my $total_msgs = scalar(@msg);
        if ($more_than_max) {
            $context->{note} = $loc->(
                'More than [_1] results were returned. ' .
                'Newest [_2] showing.', $total_msgs, $total_msgs);
        } else {
            $context->{note} = $loc->(
                'Your search returned [quant,_1,result,results].', $total_msgs);
        }
        $context->{search}{results} = \@msg;
    }
}

sub _earliest_datetime {
    my $time = shift;
    return strftime('%Y-%m-%d 00:00', localtime($time));
}

sub _latest_datetime {
    my $time = shift;
    $time += (60 * 60* 24); # to get next day's date
    return strftime('%Y-%m-%d 00:00', localtime($time));
}

sub guess_date {
  my ($date,$end_time) = @_;
  return unless $date;
  my ($date_part,$time_part);

  # contruct a local time string from the input and our defaults
  # format is 2007-12-30 14:55 , time is optional
  if ( $date =~ /(^\s*(\d{4}-\d{2}-\d{2}))/ ) {
  
      $date_part = $1;
      if ($date =~ /[\s_]+(\d\d:\d\d)$/) {
        $time_part = $1;
      } elsif($date =~ /[\s_\[](.+)$/) {
        # there was some content for a time but it was wrong :(
        return (undef,undef);
      } else {
        $time_part = $end_time ? '23:59' : "00:00";
      }
 
       
  } else {
     return;
  }
  my $local_time_string = "${date_part}T$time_part:00";
  my $display_local_time_string = "$date_part $time_part";
 
  # convert the local time string to utc for searching
  my $utc_time_string = '';
  my $raw_year = 0;
  eval { 
    my @strp = strptime($local_time_string);
    my @gm_time =  gmtime( timelocal( @strp ) ); 
    $utc_time_string = strftime('%Y-%m-%dT%H:%M:%S', @gm_time);
    $strp[(@strp-1)] = -1 if !defined $strp[(@strp-1)]; #if timzone is undef
    $raw_year = strftime('%Y', @strp);
  };
  if ($@ or $raw_year < 1972) { # anything before unix epoch is suspect
    return(undef,undef); 
  }
  return ($utc_time_string,$display_local_time_string);
}

sub format_epoch {
  my $time = shift;
  # return a local tz time string formatted in the exact same format as users input
  return strftime('%Y-%m-%d %H:%M', localtime($time));
}

sub handle_logsearch_preview {
    my ($app) = @_;
    my $context = $app->{context};
    $context->{tab} = 'logsearch';
    $context->{page} = 'details';
    
    my $user = $context->{user};
    my $groups;
    if(!$user->super_user){
        # XXX: hack to overcome cdbi's stupid use of overload
        $groups = [
#            map { '' . $_->group_name } $user->groups
             map { my $g = '' . $_->group_name; $g =~ s/_/__/g; $g } $user->groups
        ];
    }

    my $key = $app->query->path_info();
    $key =~ s{.*/([^/]+)$}{$1};
    my $ls = Sophos::LogSearch->new();
    my $msg;

    my $logs = {};
    $logs->{"no_message_log"} = 1 
        unless $user->any_group_can("logsearch.preview.rawlogs.message_log");
    
    eval { $msg = $ls->message($key, $groups, $logs) };
    
    if (!$msg || (my $errors = $ls->http_errors()) || $@) {
        $app->log("Unable to retrieve message details from remote host.\n");
        $app->log("\tDied: $@") if $@;
        for my $error (@$errors){
            $app->log("\t$error->{url} --> $error->{message}\n");
        }

        die $context->{loc}->("Unable to retrieve message details from remote host\n");
    }

    my $message = $context->{msg} = {};
    $message->{rawlogs}{maillog_incoming} = $msg->maillog_incoming();
    $message->{rawlogs}{maillog_outgoing} = $msg->maillog_outgoing();

    $message->{rawlogs}{messagelog} = '';
    # Don't search locally if the user does not have permission
    unless ($logs->{no_message_log}) {
        chomp($message->{rawlogs}{messagelog} = $msg->message_log_data() || '');
    }
    $message->{splits} = $msg->split_data();
    
    $message->{logdata} = $msg->logdata();
    $message->{milterhost} = $msg->milter_host() || '';

    for my $field (qw/subject from to time client relay msgid action reason/) {
        if ($field eq 'time') {
            # convert time from UTC
            eval {
                my @strp = strptime($msg->field($field));
                my @local_time = localtime( timegm(@strp) );
                my $local_time_string = strftime('%Y-%m-%dT%H:%M:%S', @local_time);
                $message->{$field} = $local_time_string;
            };
            if ($@) {
                $message->{$field} = $msg->field($field);
            }
        } else {
            $message->{$field} = $msg->field($field);
        }
    }
    if (my $routes = $msg->field('routes')) {
        $message->{routes} = join(', ', @$routes);
    }
    if (my $copies = $msg->field('copies')) {
        $message->{copies} = join(', ', @$copies);
    }

    $message->{time} =~ s{T}{ };
    $message->{subject} =~ s{\^\^}{ }g;
    $message->{to} = join(', ', split(',', $message->{to}));
    $message->{action} = join(', ', split(',', $message->{action}));
    $app->log("logsearch: " . "Viewing message 'info' for message '$key'");
}
#chankey pathak: testing something
1;
