package PureMessage::WebUI::Controller::Message;
use strict;
use warnings;
use PureMessage::Quarantine;
use PureMessage::QMan qw(full_clean_body full_message_source 
                         full_attachments_info fetch_attachment
                         full_message_minfo);
use base 'Exporter';
our @EXPORT_OK = qw(message_details print_message_attachment forward_message show_original);
use PureMessage::WebUI::Utils qw(htmlify get_message_status
                                 calc_message_status);
use LWP::UserAgent;
use Apache::Constants qw(FORBIDDEN REDIRECT);
use ActiveState::I18N::UTF8 qw(utf8_header);
use PureMessage::UTF8::Config qw(%utf8_config);
use PureMessage::Config qw(%config);
use PureMessage::WebUI::RunModeException;
use PureMessage::Group::Forward;
use PureMessage::Install::OptionValidator qw/valid_email/;
use PureMessage::Queue qw(queue_mail);
use MIME::Entity;

# load the permissions for the message tabs
sub load_message_perms {
    my ($app, $msg, $page) = @_;

    my $context = $app->{context};
    my $user = $context->{user};

    my $user_groups = [ map { ''.$_->group_name } $user->groups ];
    my $msg_groups = $msg->message_groups();

    # Allow superusers to calculate permissions for messages
    # with no groups
    $msg_groups = ['supergroup'] if $user->super_user and !@$msg_groups;

    for my $g (@$msg_groups) {
        # Skip if current message group is not administered by user
        next unless grep { $_ eq $g } @$user_groups;

        # Go through each quarantine.preview permission for current message
        # group
        for my $perm (qw(content source info attachments status
                         attachments.download)) {
            # If we already have seen a true setting for a permission,
            # skip onto the next one.
            next if $context->{perm}{message}{$perm};

            my $yes = $user->can_do(
                group => $g, perm => qq{quarantine.preview.$perm}
            );
            if ($perm eq 'attachments.download') {
                $context->{perm}{actions}{download_attachment} = $yes ? 1 : 0;
            }
            else {
                $context->{perm}{message}{$perm} = $yes ? 1 : 0;
            }
        }
    }

    unless($context->{perm}{message}{$page}) {
        my $redirect_to = q{};
        my @paging_order = qw(content source info attachments status);
        for my $redirect_page (@paging_order){
            next unless $context->{perm}{message}{$redirect_page};
            $redirect_to = $redirect_page;
            last;
        }

        my $message_id = $msg->id();

        $app->log(qq{Access to page '$page' for '$message_id' prohibited});
        if ($redirect_to) {
            die PureMessage::WebUI::MessagePreviewException->new(
                    redirect => qq{/groups/message/$redirect_to/$message_id});
        }
        else {
            die PureMessage::WebUI::MessagePreviewException->new(
                    prohibited => 1);
        }
    }
}

sub message_details {
    my ($app, $page, $msg_id) = @_;

    my $context = $app->{context};
    $context->{tab} = q{message};
    $context->{page} = q{default};
    $context->{message} = {msg_id => $msg_id};

    # check for invalid pages
    my %valid_page = map { $_ => 1 }
                      qw(content status info attachments source);
    # XXX localize error message
    die PureMessage::WebUI::MessagePreviewException->new(
              error_msg => qq{$page isn't a valid message page!\n})
        unless $valid_page{$page};

    $context->{page} = $page;
    $context->{page_name} = $context->{loc}->(ucfirst($page));

    my $msg = _load_message_object($msg_id);

    load_message_perms($app, $msg, $page);

    my $from = $msg->header('h_from') || '';
    my $to   = $msg->header('h_to') || '';
    my $subj = $msg->c_subject_utf8 || '';
    utf8_header($to, \%utf8_config),
    utf8_header($from, \%utf8_config),
    utf8_header($subj, \%utf8_config),

    my %msg_details = (
        msg_id => $msg_id,
        to      => $to,
        from    => $from,
        subject => $subj,
    );

    $app->log(qq{Viewing message '$page' for message id '$msg_id'});

    # XXX Always include attachment info so we can disable the menu if
    # XXX necessary. This is currently a very expensive operation. What
    # XXX needs to happen, is to allow fast retrieval of attachment
    # XXX metadata, rather than the message being rescanned each time
    # XXX we want to get at it. We should never (give or take) have to
    # XXX _rescan_ a message. We should have all metadata in the minfo.
    $msg_details{attach_info} = full_attachments_info($msg);

    if ($page eq q{content}) {
        $msg_details{body} = full_clean_body($msg);
    }
    elsif ($page eq q{info}) {
        # XXX Get the virus id from disk since we should not store it in db.
        my @dir_minfo = @{ full_message_minfo($msg) };

        $msg_details{minfo} = [ $msg->minfo_items() ];
        while(@dir_minfo) {
            my ($key, $value) = splice(@dir_minfo, 0, 2);
            if ($key eq 'Virus-Id' || $key eq 'Direction' || $key eq 'Spam-Hit') {
                push @{$msg_details{minfo}}, ($key, $value);
            } elsif ( $key eq 'Matched-Word' || $key eq 'Matched-Attachment') {
              my ($i) = grep ${$msg_details{minfo}}[$_] eq 'Reason', 0..@{$msg_details{minfo}};
              ${$msg_details{minfo}}[$i+1] .= " ( $value )";
            }
        }
    }
    elsif ($page eq q{status}) {
        my $status = get_message_status($msg);
        $msg_details{status} = calc_message_status($status);
        $msg_details{recipients} = recipient_status($app, $msg);
    }
    elsif ($page eq q{source}){
        $msg_details{source} = full_message_source($msg);
    }

    # Make sure that HTML Entities are escaped properly in all of the
    # info items.
    foreach my $key (keys %msg_details) {
        $msg_details{$key} = htmlify($msg_details{$key} || '');
    }

    $context->{message} = \%msg_details;
}

sub _load_quarantine_object {
    return PureMessage::Quarantine->new();
}

sub _load_message_object {
    my ($msg_id, $mstore) = @_;
    $mstore ||= _load_quarantine_object();
    my $msg = $mstore->message($msg_id);

    # XXX localize error message
    die PureMessage::WebUI::MessagePreviewException->new(
              error_msg => qq{Couldn't fetch message ($msg_id)\n})
        unless $msg;
    return $msg;
}

sub recipient_status {
    my ($app, $msg) = @_;

    my @recips = $msg->minfo_value('Envelope-To');
    my %actions = $msg->user_actions(\@recips);
    my $i = 0;
    for my $r (@recips) {
        my $action = $actions{$r} || '';
        my $status = { approve => 'approved',
                       delete  => 'deleted',
                     }->{$action} || 'empty';
        my $status_desc = $status eq 'empty' ? 'No action' : $status;
        $r = { num => $i++,
               address => $r,
               status => $status,
               status_desc => $app->{context}{loc}->(ucfirst $status_desc),
             };
    }
    return \@recips;
}

sub user_message_can {
    my ($user, $msg, $perm) = @_;
    my $user_groups = [ map { ''.$_->group_name } $user->groups ];
    my $msg_groups = $msg->message_groups();
    my $can = 0;

    # Super users can administer all messages, even ones that don't 
    # have any groups
    return $user->can_do(perm => $perm) if $user->super_user;

    for my $g (@$msg_groups) {
        # Skip if current message group is not administered by user
        next unless grep { $_ eq $g } @$user_groups;
        next unless $user->can_do(group => $g, perm => $perm);
        $can++;
        last;
    }
    return $can;
}

sub print_message_attachment {
    my ($app, $msg_id, $part_id) = @_;

    # check that user can download attachments for THIS PARTICULAR message
    my $user = $app->{context}{user};
    my $msg = _load_message_object($msg_id);
    my $can_download = user_message_can($user, $msg, 
                            'quarantine.preview.attachments.download');
    unless ($can_download) {
        $app->log("No permission to download attachment $part_id for message "
                  . "$msg_id");
        die PureMessage::WebUI::MessagePreviewException->new(prohibited => 1);
    }

    $app->log("Downloading attachment $part_id for message $msg_id");

    my $printed_headers = 0;
    my $callback = sub {
        my ($data, $resp) = @_;
        unless ($printed_headers) {
            $printed_headers++;
            my $headers = $resp->headers;
            my $size = $headers->header('Content-Length');
            my $cd = $headers->header('Content-Disposition');
            my $type = $headers->header('Content-Type') || 'text/plain';

            # tell cgi-app not to output headers, then do it ourself
            $app->header_type('none');
            print $app->query->header(
                '-type' => $type,
                '-Content-Disposition' => $cd,
                '-Content-Length' => $size,
            );
        }
        unless (print $data) {
            $app->log("Connection lost, aborting download of $msg_id:$part_id");
            die 'Hey LWP: cancel the download';
        }
    };

    fetch_attachment($msg, $part_id, $callback);
}

sub show_original {
    my ($app, $msg_id) = @_;
    die "Cannot fetch message ID for showing original" unless $msg_id;

    my $user = $app->{context}{user};
    my $msg = _load_message_object($msg_id);
    my $can = user_message_can($user, $msg, 'quarantine.preview.source');
    unless($can) {
        $app->log("No permission to view source for message: $msg_id");
        die PureMessage::WebUI::MessagePreviewException->new(prohibited => 1);
    }

    $app->log("Show original for message $msg_id");
    return full_message_source($msg);
}

sub forward_message {
    my ($app) = @_;
    my $context = $app->{context};
    my $user = $context->{user};
    my $query = $app->query;

    my $msg_id_str = $query->param('msg_ids') || '';
    $context->{msg_ids} = $msg_id_str;
    my @msg_ids = split ',', $msg_id_str;
    unless (@msg_ids) {
        die "No messages selected to forward.\n";
    }
    
    $app->log(
        $context->{loc}->(
            q{Attempting to forward [quant,_1,message,messages]},
            scalar(@msg_ids)
        )
    );

    my $q    = _load_quarantine_object();
    my @msgs = map { _load_message_object($_, $q) } @msg_ids;

    my $type = $query->param('forward_type');
    if (!$type) {
        # Show the form
        _load_forward_configs($context, $user, \@msgs);
        return;
    }
    my $perm = "quarantine.actions.forward.$type";

    my @forward_addresses;
    my $comment   = $query->param('comments') || '';
    if ($type eq 'custom') {
        # Find and validate the custom email addresses
        my $fwd_addrs = $query->param('fwd_addresses');
        eval {
            unless ($fwd_addrs) {
                $app->log(qq{Attempted to forward 'custom' without addresses});
                die $context->{loc}->("Please supply an address.\n");
            }
            for my $email (split /\r?\n/, $fwd_addrs) {
                eval { valid_email($email) };
                if($@){
                    $app->log(qq{Attempted to forward 'custom' with bad address});
                    die $context->{loc}->("'[_1]' is not a valid address\n", $email);
                }
                push @forward_addresses, $email;
            }
        };
        if ($@) {
            $context->{fwd_error} = $@;
            $context->{fwd_default} = 'custom';
            $context->{fwd_comment} = $comment;
            _load_forward_configs($context, $user, \@msgs);
            return;
        }
    }
    else {
        my $fwd = PureMessage::Group::Forward->search(access_right => $perm)->first;
        push @forward_addresses, $fwd->address;
    }

    my $to_addr = join(q{, }, @forward_addresses);
    
    # Determine which messages we can forward
    my @can_forward;
    for my $msg (@msgs) {
        if (user_message_can($user, $msg, $perm)) {
            push @can_forward, $msg;
            
            my $reasoning = $type eq q{custom} 
                ? qq{(custom - $to_addr)}
                : qq{($type)};

            $app->log(qq{Forwarded message }.$msg->id().qq{ $reasoning});
        }
        else {
            $app->log("No permission to forward message ".$msg->id()." ($type)");
        }
    }

    # forward the message
    my $subject = $type eq q{custom} 
        ? $context->{loc}->(q{FW:  PureMessage Quarantined Messages})
        : PureMessage::Group::Forward
            ->search(access_right => $perm)
            ->first()
            ->subject();

    my $from_addr = $user->email();

    my $top = MIME::Entity->build(
        Type    => q{multipart/mixed},
        To      => $to_addr,
        From    => $from_addr,
        Subject => "$subject",
    );

    $top->attach(
        Data => $comment,
        Type => q{text/plain},
        Encoding => q{quoted-printable},
        Charset  => q{UTF-8},
    );

    for my $msg (@can_forward){
        my $source = full_message_source($msg);
        $top->attach(
            Data        => $source,
            Encoding    => q{8bit},
            Filename    => $msg->id() . q{.eml},
            Disposition => q{attachment},
            Type        => q{message/rfc822},
        );
    }

    eval {
        my $tmp = $q->new_tempfile("forward-$$");
        my $tmp_filename = $tmp->filename;
        open my $fh,  ">$tmp_filename" or 
            die("Cannot create $tmp_filename: $!\n");
        $top->print($fh);
        close $fh;

        queue_mail(
           to       => [ @forward_addresses ],
           from     => $from_addr,
           tempfile => $tmp,
        );
    };

    if ($@) {
        (my $err = $@) =~ s/\n/ /sg;
        $app->log("Unable to queue message: $err");        
        die PureMessage::WebUI::MessagePreviewException->new(
                  error_msg => qq{Unable to queue message for delivery\n});
    }

    # Show the results of the operation
    my $address = join ', ', @forward_addresses;
    my $msg = $context->{loc}->(
        (
            ( scalar(@msg_ids) > 1 and !(scalar(@msg_ids) == scalar(@can_forward)) )
            ? 
                ("Forwarded [_1] / [quant,_2,message,messages] to [_3].",
                    scalar(@can_forward), scalar(@msg_ids), $address)
            :
                ("Forwarded [quant,_1,message,messages] to [_2].",
                    scalar(@can_forward), $address)
        )
    );
    $context->{result_msg} = $msg;
}

sub _load_forward_configs {
    my ($context, $user, $msgs) = @_;
    my @configs = PureMessage::Group::Forward->retrieve_all;
    for my $c (@configs) {
        my $can;
        my $perm = '' . $c->access_right; # hack for Class::DBI
        for my $msg (@$msgs) {
            $can++ if user_message_can($user, $msg, $perm);
        }
        if ($can) {
            push @{ $context->{forward_configs} }, $c;
            $context->{fwd_default} ||= $c->name;
        }
    }

    # Check out the custom forward permission
    for my $msg (@$msgs) {
        if (user_message_can($user, $msg, 'quarantine.actions.forward.custom')) {
            $context->{forward_custom} = 1;
            last;
        }
    }
}

1;
